

Imports System.Data

Namespace DBSaber

    ''' <summary>
    ''' This class contains common database tools used when working with SQL databases. 
    ''' They are as general as possible but certain things like SQL Parameters are SQL Server/Access Specific
    ''' Cuts through databases like a Light Saber.
    ''' Assumptions: No NULL VALUES IN THE DATABASE, SQL Server or Access
    ''' </summary>
    Public Class StdTools


        ''' <summary>
        ''' The format to use when converting date objects to string for use in SQL statements.
        ''' </summary>
        Public Const cFDate As String = "{0:MM-dd-yyyy HH:mm:ss}"


        ''' <summary>
        ''' *Database should be set up with this as the default date
        ''' The NullDate time is 00:00:00
        ''' </summary>
        Public Const cNullDate As Date = #1/1/1900#
        Public Const cMinDate As Date = #1/1/1900#
        Public Const cMaxDate As Date = #1/1/2112#


#Region " SQL Tools "

        ''' <summary>
        ''' Fixes strings for use in SQL statements that might have invalid characters
        ''' </summary>
        Public Shared Function SQLStrClean(ByVal vStr As String) As String

            Return vStr.Replace("'", "''")

        End Function

        ''' <summary>
        ''' Takes a string and make it LIKE friendly by putting ' and wild cards (%) at both ends
        ''' Cleans the string for invalid characters
        ''' </summary>
        Public Shared Function SQLStrLike(ByVal vStr As String) As String

            Return "'%" & SQLStrClean(vStr) & "%'"

        End Function


#Region " SQLValue Functions "

        ''' <summary>
        ''' Creates an SQL Friendly value of the String data type. 
        ''' String have to be escaped. 
        ''' Usage: FieldName=" SQLValue
        ''' </summary>
        Public Shared Function SQLValue(ByVal value As String) As String
            'Need to clean up the strings. Helps prevent SQL injection attacks
            Return "'" & StdTools.SQLStrClean(value) & "'"
        End Function

        'Strings have to have escaped. Usage: WHERE PartName= & SQLValue
        'Creates an SQL Friendly value of the String data type. 

        ''' <summary>
        ''' Creates an SQL Friendly value of the Date data type.
        ''' Database ShortDate fields will round the seconds to the nearest minute on entry into the Database.
        ''' </summary>
        Public Shared Function SQLValue(ByVal value As Date) As String
            Return "'" & String.Format(cFDate, value) & "'"
        End Function

        ''' <summary>
        ''' Creates an SQL Friendly value of the Decimal data type.
        ''' </summary>
        Public Shared Function SQLValue(ByVal value As Decimal) As String
            Return value.ToString
        End Function

        ''' <summary>
        ''' Creates an SQL Friendly value of the Single data type.
        ''' </summary>
        Public Shared Function SQLValue(ByVal value As Single) As String
            Return value.ToString
        End Function

        ''' <summary>
        ''' Creates an SQL Friendly value of the Date data type.
        ''' </summary>
        Public Shared Function SQLValue(ByVal value As Int32) As String
            Return value.ToString
        End Function

        ''' <summary>
        ''' Creates an SQL Friendly value of the Int16 data type.
        ''' </summary>
        Public Shared Function SQLValue(ByVal value As Int16) As String
            Return value.ToString
        End Function

        ''' <summary>
        ''' Creates an SQL Friendly value of the Boolean data type.
        ''' Output: "1" or "0"
        ''' </summary>
        Public Shared Function SQLValue(ByVal value As Boolean) As String

            If value = True Then
                Return "1"
            Else
                Return "0"
            End If

        End Function

#End Region


#End Region


#Region " Date Support "

        ''' <summary>
        ''' Creates a specified date with the earliest time possible (12:00:00 am)
        ''' </summary>
        ''' <param name="vDate"></param>
        ''' <returns>Date with the earliest time string possible</returns>
        ''' <remarks></remarks>
        Public Shared Function DateTimeEarliest(ByVal vDate As Date) As Date

            'Using String.Format to create the datestring with the proper time
            '
            'Note: There probably is a better way to do this because the string
            '      parsing is most likely slow. I could do raw math using the addhours,mins, seconds
            '      methods but that might be slow as well. It is wrapped up in this function
            '      so I could always change it out later.

            Return CType(String.Format(String.Format("{0:MM/dd/yyyy 00:00:00}", vDate)), Date)

        End Function

        ''' <summary>
        ''' Used for sql queries
        ''' </summary>
        ''' <param name="vDate"></param>
        ''' <returns></returns>
        ''' <remarks></remarks>
        Public Shared Function DateTimeEarliestStr(ByVal vDate) As String

            Return String.Format(String.Format("{0:MM/dd/yyyy 00:00:00}", vDate))
        End Function


        ''' <summary>
        ''' Creates a specified date with the latest time string possible (11:59:59 pm)
        ''' </summary>
        ''' <param name="vDate"></param>
        ''' <returns>Date with the latest time string possible</returns>
        ''' <remarks></remarks>
        Public Shared Function DateTimeLatest(ByVal vDate As Date) As Date

            'Note: See DateTimeEarliest for notes.
            Return CType(String.Format("{0:MM/dd/yyyy 23:59:59}", vDate), Date)

        End Function


        Public Shared Function DateTimeLatestStr(ByVal vDate As Date) As String

            'Note: See DateTimeEarliest for notes.
            Return String.Format("{0:MM/dd/yyyy 23:59:59}", vDate)

        End Function

        'Purpose: Determines the first day of the week
        '
        'Note: Didn't want to use the weekday function but couldn't find a better way

        ''' <summary>
        ''' Determines the first day of the week (sunday) for a specified date
        ''' </summary>
        ''' <param name="vDate">The date to bump to the first day of the week</param>
        ''' <param name="vMakeEarliestTime">Whether or not to adjust the date's time to the earliest time possible (12:00:00 am)</param>
        ''' <returns>The date that is the first day of the week</returns>
        ''' <remarks></remarks>
        Public Shared Function DateWeekStart(ByVal vDate As Date, Optional ByVal vMakeEarliestTime As Boolean = False) As Date

            Dim tDate As Date = vDate.AddDays(-1 * (Weekday(vDate) - 1))

            If vMakeEarliestTime = True Then
                'They want the earliest time 00:00:00
                Return DateTimeEarliest(tDate)
            Else
                'They want the time they already had in the date
                Return tDate
            End If

        End Function

        ''' <summary>
        ''' Determines the Last day of the week (Saturday's date) for a specified date
        ''' </summary>
        ''' <param name="vDate"></param>
        ''' <param name="vMakeLatestTime"></param>
        ''' <returns></returns>
        ''' <remarks></remarks>

        Public Shared Function DateWeekEnd(ByVal vDate As Date, Optional ByVal vMakeLatestTime As Boolean = False) As Date

            Dim tDate As Date = vDate.AddDays(7 - Weekday(vDate))

            If vMakeLatestTime = True Then
                'They want the earliest tie 23:59:59
                Return DateTimeLatest(tDate)
            Else
                'They wan the time the already had in the date
                Return tDate
            End If

        End Function

        ''' <summary>
        ''' Creates a date time with the 1/1/xxxx 00:00:00
        ''' </summary>
        ''' <param name="vYear"></param>
        ''' <returns></returns>
        ''' <remarks></remarks>
        Public Shared Function DateYearStart(ByVal vYear As Integer) As Date

            Return Date.Parse("1/1/" & vYear & " 00:00:00")

        End Function

        ''' <summary>
        ''' Creates a date with the last day of the year 12/31/xxxx 23:59:59
        ''' </summary>
        ''' <param name="vYear"></param>
        ''' <returns></returns>
        ''' <remarks></remarks>
        Public Shared Function DateYearEnd(ByVal vYear As Integer) As Date


            Return Date.Parse("12/31/" & vYear & " 23:59:59")
            'Return CType(String.Format("{0:12/31/" & vYear & " 23:59:59}", vDate), Date)

        End Function


        ''' <summary>
        ''' Compares 2 dates down to the 100th of a second or somewhere around there. Had
        ''' a problem with the ticks not being exact after reading in a db value.
        ''' </summary>
        ''' <param name="vDate1"></param>
        ''' <param name="vDate2"></param>
        ''' <returns></returns>
        ''' <remarks>The issue involves setting a date to now, writing the date to teh SQLServer
        '''          then reading the date back from the sql server. the ticks aren't exact but very close.
        '''          I suspect there is a rounding issue. I thought this might solve the problem.
        '''   Write: 632774639849695264 compared to read: 632774639849700000
        ''' </remarks>
        Public Shared Function DateLooseEquals(ByVal vDate1 As Date, ByVal vDate2 As Date) As Boolean

            Const cF As String = "{0:MM-dd-yyyy hh:mm:ss:ff}"

            'SHould probably go down to the millisecond.

            'Improve: There is probably a better way to compare dates to the millisecond
            '         then converting to strings.

            If String.Compare(String.Format(cF, vDate1), String.Format(cF, vDate2)) = 0 Then
                Return True
            End If

        End Function


        Public Shared Function DateDayEquals(ByVal vDate1 As Date, ByVal vDate2 As Date) As Boolean

        End Function

#End Region


#Region " Command Tools "

        ''' <summary>
        ''' Adds a parameter and value to a DbCommand.
        ''' *Be careful with the parameter orders because in some situations it matters
        ''' </summary>
        ''' <param name="rCommand">The DbCommand that needs a parameter</param>
        ''' <param name="vParamName">The name of the SQL parameter to add</param>
        ''' <param name="vDBType">The SQL Parameter's Data Type (String, Int32, Date, etc.)</param>
        ''' <param name="vValue">The SQL Parameter Value</param>
        ''' <remarks></remarks>
        Public Shared Sub CommandParamAdd(ByRef rCommand As Data.Common.DbCommand, ByVal vParamName As String, ByVal vDBType As Data.DbType, ByVal vValue As Object)

            'Creating the parameter of the Command
            Dim tParam As Data.Common.DbParameter = rCommand.CreateParameter

            tParam.ParameterName = vParamName 'Setting the parameter
            tParam.DbType = vDBType 'Setting the specified datatype
            tParam.Value = vValue

            rCommand.Parameters.Add(tParam)


        End Sub


#End Region


#Region " DataReader Tools "

        ''' <summary>
        ''' Retrieves a single record's field value from the database using Parametized SQL.
        ''' If multiple Fields are specified in the SQL the first value is retrieved. THROWS Exception if record not found
        ''' </summary>
        Public Shared Function DBValue(ByRef rConn As Data.Common.DbConnection, ByVal vSQL As String, ByVal vParamName As String, ByVal vParamDBType As DbType, ByVal vParamValue As Object) As Object

            Dim tCom As Data.Common.DbCommand
            Dim DR As Data.Common.DbDataReader


            'Creating the command
            tCom = rConn.CreateCommand
            tCom.CommandText = vSQL

            CommandParamAdd(tCom, vParamName, vParamDBType, vParamValue)

            DR = tCom.ExecuteReader

            If DR.HasRows = True Then
                Call DR.Read()
                DBValue = DR.GetValue(0)
                Call DR.Close()
            Else
                'Didn't find the value
                Call DR.Close()
                Throw New NoRecordFoundDBException(vParamName, vParamValue)
            End If

        End Function


        ''' <summary>
        ''' Retrieves a single record's field value from the database using Pure SQl
        ''' If multiple Fields are specified in the SQL the first value is retrieved.
        ''' </summary>
        Public Shared Function DBValue(ByRef rConn As Data.Common.DbConnection, ByVal vSQL As String) As Object

            Dim tCom As Data.Common.DbCommand
            Dim DR As Data.Common.DbDataReader


            'Creating the command
            tCom = rConn.CreateCommand
            tCom.CommandText = vSQL

            DR = tCom.ExecuteReader

            If DR.HasRows = True Then
                Call DR.Read()
                DBValue = DR.GetValue(0)
                Call DR.Close()
            Else
                'Didn't find the value
                Call DR.Close()
                Throw New NoRecordFoundDBException(vSQL)
            End If

        End Function

        ''' <summary>
        ''' Retrieves records from the database and puts them in a collection
        ''' If multiple Fields are specified in the SQL the first value is retrieved. Throws an error if no records found
        ''' </summary>
        ''' <returns>A collection of Database values</returns>
        Public Shared Function DBCollect(ByRef rConn As Data.Common.DbConnection, ByVal vSQL As String, Optional ByVal vThrowNoFoundError As Boolean = True) As Collection

            Dim tCom As Data.Common.DbCommand
            Dim DR As Data.Common.DbDataReader
            Dim tCollect As Collection

            'Creating the command
            tCom = rConn.CreateCommand
            tCom.CommandText = vSQL

            DR = tCom.ExecuteReader

            If DR.HasRows = True Then
                tCollect = New Collection

                Do While DR.Read = True

                    tCollect.Add(DR.GetValue(0))

                Loop

                Call DR.Close()
                Return tCollect
            Else
                'Didn't find the value
                Call DR.Close()

                DBCollect = Nothing

                If vThrowNoFoundError = True Then
                    'They want to throw an error if there aren't any records found
                    Throw New NoRecordFoundDBException(vSQL)
                End If

            End If

        End Function


        ''' <summary>
        ''' Creates a DataReader from just SQL.
        ''' </summary>
        ''' <param name="rConn"></param>
        ''' <param name="vSQL"></param>
        ''' <returns></returns>
        ''' <remarks>This is very similar to how ado used to work back in the vb6 days. Basically I just wanted a quick way to get a datareader full of information and didn't want to deal with commands everywhere.</remarks>
        Public Shared Function DRCreate(ByRef rConn As Data.Common.DbConnection, ByVal vSQL As String) As Data.Common.DbDataReader

            Dim DR As Data.Common.DbDataReader
            Dim tCommand As Data.Common.DbCommand = rConn.CreateCommand

            tCommand.CommandText = vSQL
            DR = tCommand.ExecuteReader

            Return DR

        End Function
        ''' <summary>
        ''' Creates a DataReader from just SQL.
        ''' </summary>
        ''' <remarks>Used in situation like i row selects</remarks>
        Public Shared Function DRCreate(ByRef rConn As Data.Common.DbConnection, ByVal vSQL As String, ByVal vParamName As String, ByVal vDBType As DbType, ByVal vValue As Object) As Data.Common.DbDataReader

            Dim DR As Data.Common.DbDataReader
            Dim tCommand As Data.Common.DbCommand = rConn.CreateCommand

            tCommand.CommandText = vSQL


            CommandParamAdd(tCommand, vParamName, vDBType, vValue)

            DR = tCommand.ExecuteReader

            Return DR

        End Function

        ''' <summary>
        ''' Executes a NonQuery sql against the DB Connection
        ''' </summary>
        Public Shared Sub SQLExec(ByRef rConn As Data.Common.DbConnection, ByVal vSQL As String)

            Dim Cmd As Data.Common.DbCommand = rConn.CreateCommand

            Cmd.CommandText = vSQL

            Call Cmd.ExecuteNonQuery()

        End Sub

#End Region


        'Public Shared Sub LostUpCheck(ByRef rConn As Data.Common.DbConnection, ByVal vTableName As String, ByVal vRecDate As Date, ByVal vParamName)


        '    Dim tRecLastModified As Date = DBSaber.StdTools.DBValue(rConn, "SELECT LastModified FROM " & vTableName & " WHERE VesselSpecID=@pVesselSpecID", "pVesselSpecID", DbType.Int32, Me.VesselSpecID)

        '    If tRecLastModified <> vLastModified Then
        '        Throw New DBSaber.LostUpdateDBException(tRecLastModified, Me.LastModified)
        '    End If


        'End Sub

        ''' <summary>
        ''' Safely closes a database connection, even if is nothing
        ''' </summary>
        Public Shared Sub ConnClean(ByRef rConn As Data.Common.DbConnection)

            If Not (rConn Is Nothing) Then
                Call rConn.Close()
            End If

        End Sub


        'Purpose: Takes a collection of strings or numbers and creates an IN statement
        '
        'Note: Need to prefix the IN statement with a Field Name
        '
        'Note: IN ('Cheese','Coke','Burger')
        Public Shared Function SQLIN(ByRef rCollect As Collection) As String

            Dim tObj As Object
            Dim tStr As String = " IN ("


            If rCollect.Count = 0 Then
                Throw New Exception("SQLIn collection empty")
            End If


            'ToDo: Make this work with different data types
            For Each tObj In rCollect
                'Creating the IN string
                tStr = SQLStrClean(tStr) & (tObj) & ","

            Next

            'Removing the last ","
            If tStr.EndsWith(",") Then
                tStr = tStr.Substring(0, tStr.Length - 1)
            End If


            Return tStr & ")"

        End Function


        ''' <summary>
        ''' Retrieves a string from the datareader. "" if null
        ''' </summary>
        ''' <param name="rDR"></param>
        ''' <param name="oridinal"></param>
        ''' <returns></returns>
        ''' <remarks></remarks>
        Public Shared Function DBValStr(ByRef rDR As Data.Common.DbDataReader, ByVal oridinal As Integer) As String

            Dim blah As Object

            blah = rDR.GetValue(oridinal)

            If TypeOf blah Is System.DBNull Then
                Return ""
            Else
                Return CType(blah, String)
            End If


        End Function

        ''' <summary>
        ''' retrieves a decimal from the datareader. 0 if null
        ''' </summary>
        ''' <param name="rDR"></param>
        ''' <param name="oridinal"></param>
        ''' <returns></returns>
        ''' <remarks></remarks>
        Public Shared Function DBValDec(ByRef rDR As Data.Common.DbDataReader, ByVal oridinal As Integer) As Decimal

            Dim blah As Object

            blah = rDR.GetValue(oridinal)

            If TypeOf blah Is System.DBNull Then
                Return 0
            Else
                Return CType(blah, Decimal)
            End If


        End Function

        ''' <summary>
        ''' retrieves a date value from the datareader. NullDate if null
        ''' </summary>
        ''' <param name="rDR"></param>
        ''' <param name="oridinal"></param>
        ''' <returns></returns>
        ''' <remarks></remarks>
        Public Shared Function DBValDate(ByRef rDR As Data.Common.DbDataReader, ByVal oridinal As Integer) As Date

            Dim blah As Object

            blah = rDR.GetValue(oridinal)

            If TypeOf blah Is System.DBNull Then
                Return DBSaber.StdTools.cNullDate
            Else
                Return CType(blah, Date)
            End If


        End Function

    End Class


#Region " Exceptions "

    ''' <summary>
    ''' This is the standard exception when a database record cannot be found in a table.
    ''' </summary>
    <System.Serializable()> Public Class NoRecordFoundDBException

        Inherits System.Exception

        'Public Sub New()
        'End Sub 'New

        'Public Sub New(ByVal message As String)
        '    MyBase.New(message)
        'End Sub 'New

        'Public Sub New(ByVal message As String, ByVal inner As Exception)
        '    MyBase.New(message, inner)
        'End Sub 'New


        ''' <summary>
        ''' Used to provide detailed information about the keys and the table for which there is no record.
        ''' </summary>
        Public Sub New(ByVal vDBTable As String, ByVal vKeyName As String, ByVal vKeyValue As String)

            MyBase.New("Couldn't find " & vKeyName & "=" & vKeyValue & " in " & vDBTable)

        End Sub

        ''' <summary>
        ''' For when the table is not known
        ''' </summary>
        Public Sub New(ByVal vKeyName As String, ByVal vKeyValue As String)

            MyBase.New("Couldn't find " & vKeyName & "=" & vKeyValue)

        End Sub

        ''' <summary>
        ''' For when only the SQL is known. This is mainly used for parameterless queries and debugging
        ''' </summary>
        Public Sub New(ByVal vSQL As String)

            MyBase.New("Couldn't find DB Record SQL=" & vSQL)

        End Sub


    End Class 'NoDBRecordFoundException


    ''' <summary>
    ''' This is the standard Lost update exception. Lost updates occur when 2 users read a record, one saves the record
    ''' then the other one does without reading the newly updated record's information. Basically both users are working
    ''' off cache and the 2nd user doesn't update the cache before his write.
    ''' </summary>
    <System.Serializable()> Public Class LostUpdateDBException

        Inherits System.Exception

        Private m_RecordLastModifiedDate As Date
        Private m_CachedDate As Date

        ''' <summary>
        ''' The record's last modified date as it is currently in the database.
        ''' </summary>
        Public ReadOnly Property RecordLastModifiedDate() As Date
            Get
                Return Me.m_RecordLastModifiedDate
            End Get
        End Property

        ''' <summary>
        ''' The users cached date. The records date when it was last read
        ''' </summary>
        Public ReadOnly Property CachedDate() As Date
            Get
                Return Me.m_CachedDate
            End Get
        End Property

        'Public Sub New()
        'End Sub 'New

        'Public Sub New(ByVal message As String)
        '    MyBase.New(message)
        'End Sub 'New

        'Public Sub New(ByVal message As String, ByVal inner As Exception)
        '    MyBase.New(message, inner)
        'End Sub 'New


        ''' <summary>
        ''' Instantiates with the important dates so they can be pulled out later easily and user to tell the end user.
        ''' </summary>
        ''' <param name="vDBLastModifiedDate">The LastModifiedDate of the record how it sits in the database.</param>
        ''' <param name="vCachedDate">The LastModifiedDate of the record when the record was last read.</param>
        ''' <remarks></remarks>
        Public Sub New(ByVal vDBLastModifiedDate As Date, ByVal vCachedDate As Date)

            MyBase.New("The database record has been modified since " & vCachedDate)


            m_CachedDate = vCachedDate


        End Sub

    End Class 'LostUpdateException


#End Region




End Namespace



